package org.simantics.databoard.serialization.impl;

import gnu.trove.TObjectIntHashMap;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.List;

import org.simantics.databoard.binding.RecordBinding;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.util.IsReferableQuery;
import org.simantics.databoard.binding.util.Result;
import org.simantics.databoard.serialization.SerializationException;
import org.simantics.databoard.serialization.Serializer;
import org.simantics.databoard.serialization.Serializer.CompositeSerializer;

public class GenericRecordSerializer extends CompositeSerializer {

	RecordBinding binding;
	public Serializer[] componentSerializers;
	Integer fixedSize;
	int minSize;
	
	/**
	 * 
	 * @param binding
	 * @param componentSerializers (optional) can be set later
	 */
	public GenericRecordSerializer(RecordBinding binding, Serializer[] componentSerializers) {		
		super( IsReferableQuery.isReferable( binding.type() ) != Result.No );
		this.binding = binding;			
		this.componentSerializers = componentSerializers;
	}
	
	@Override
	public void finalizeConstruction() {
		fixedSize = null;
		for (Serializer componentSerializer : componentSerializers) {
			minSize += componentSerializer.getMinSize();
			Integer componentFixedSize = componentSerializer.getConstantSize();
			if (componentFixedSize==null) {
				fixedSize = null;
				break;
			}
			fixedSize = fixedSize==null ? componentFixedSize : fixedSize+componentFixedSize;
		}	
	}

	@Override
	public Object deserialize(DataInput in, List<Object> identities) throws IOException {
//		assertRemainingBytes(in, minSize);
		try {
			Object[] temp = new Object[componentSerializers.length];
			for(int i=0;i<componentSerializers.length;++i)
				temp[i] = componentSerializers[i].deserialize(in, identities);
			return binding.create(temp);
		} catch (BindingException e) {
			throw new IOException( e ); 
		}
	}
	
	@Override
	public void deserialize(DataInput in, List<Object> identities, Object obj) throws IOException {
//		assertRemainingBytes(in, minSize);
		try {
			for(int i=0;i<componentSerializers.length;++i) {
				Serializer cs = componentSerializers[i];
				boolean csImmutable = binding.getComponentBinding(i).isImmutable();
				if (csImmutable) {
					Object component = cs.deserialize(in, identities);
					binding.setComponent(obj, i, component);
				} else {
					Object component = binding.getComponent(obj, i);
					cs.deserialize(in, identities, component);
					binding.setComponent(obj, i, component);
				}
			}
		} catch (BindingException e) {
			throw new IOException( e ); 
		}
	}

	@Override
	public void skip(DataInput in, List<Object> identities) throws IOException, SerializationException {
		if (fixedSize != null) {
			in.skipBytes(fixedSize);
		} else {
			for(int i=0;i<componentSerializers.length;++i)
				componentSerializers[i].skip(in, identities);				
		}
	}		
	
	@Override
	public void serialize(DataOutput out, TObjectIntHashMap<Object> identities, Object obj) throws IOException {
		try {
			for(int i=0;i<componentSerializers.length;++i)
				componentSerializers[i].serialize(out, identities, binding.getComponent(obj, i));
		} catch (BindingException e) {
			throw new IOException( e ); 
		}
	}

	@Override
	public Integer getConstantSize() {			
		return fixedSize;
	}

	@Override
	public int getSize(Object obj, TObjectIntHashMap<Object> identities) throws IOException {
		try {
			if (fixedSize!=null) return fixedSize;
			int result = 0;
			for(int i=0;i<componentSerializers.length;++i)
				result += componentSerializers[i].getSize( binding.getComponent(obj, i), identities );			
			return result;
		} catch (BindingException e) {
			throw new IOException( e ); 
		}
	}
	
	/**
	 * @return the minSize
	 */
	public int getMinSize() {
		return minSize;
	}
	
}
